# -*- coding: utf-8 -*-
r"""
Graphics arrays and insets

This module defines the classes :class:`MultiGraphics` and
:class:`GraphicsArray`. The class :class:`MultiGraphics` is the base class
for 2-dimensional graphical objects that are composed of various
:class:`~sage.plot.graphics.Graphics` objects, arranged in a given canvas.
The subclass :class:`GraphicsArray` is for
:class:`~sage.plot.graphics.Graphics` objects arranged in a regular array.

AUTHORS:

- Eric Gourgoulhon (2019-05-24): initial version, refactoring the class
  ``GraphicsArray`` that was defined in the module :mod:`~sage.plot.graphics`.

"""
from __future__ import print_function, absolute_import
import os
from sage.misc.fast_methods import WithEqualityById
from sage.structure.sage_object import SageObject
from sage.misc.temporary_file import tmp_filename
from .graphics import Graphics, ALLOWED_EXTENSIONS, _parse_figsize


class MultiGraphics(WithEqualityById, SageObject):
    r"""
    Base class for objects composed of :class:`~sage.plot.graphics.Graphics`
    objects.

    Both the display and the output to a file of ``MultiGraphics`` objects
    are governed by the method :meth:`save`, which is called by the rich output
    display manager, via
    :meth:`~sage.repl.rich_output.display_manager.DisplayManager.graphics_from_save`.

    The user interface is through the functions
    :func:`~sage.plot.plot.multi_graphics` (generic multi-graphics) and
    :func:`~sage.plot.plot.graphics_array` (subclass :class:`GraphicsArray`).

    INPUT:

    - ``graphics_list`` -- a list of graphics along with their positions on the
      common canvas; each element of ``graphics_list`` is either

      - a pair ``(graphics, position)``, where ``graphics`` is a
        :class:`~sage.plot.graphics.Graphics` object and ``position`` is the
        4-tuple ``(left, bottom, width, height)`` specifying the location and
        size of the graphics on the canvas, all quantities being in fractions
        of the canvas width and height

      - or a single :class:`~sage.plot.graphics.Graphics` object; its position
        is then assumed to occupy the whole canvas, except for some padding;
        this corresponds to the default position
        ``(left, bottom, width, height) = (0.125, 0.11, 0.775, 0.77)``

    EXAMPLES:

    A multi-graphics made from two graphics objects::

        sage: g1 = plot(sin(x^3), (x, -pi, pi))
        sage: g2 = circle((0,0), 1, color='red')
        sage: G = multi_graphics([g1, (g2, (0.2, 0.55, 0.3, 0.3))])
        sage: G
        Multigraphics with 2 elements

    .. PLOT::

        g1 = plot(sin(x**3), (x, -pi, pi))
        g2 = circle((0,0), 1, color='red')
        G = multi_graphics([g1, (g2, (0.2, 0.55, 0.3, 0.3))])
        sphinx_plot(G)

    Since no position was given for ``g1``, it occupies the whole canvas.
    Moreover, we note that ``g2`` has been drawn over ``g1`` with a white
    background. To have a transparent background instead, one has to construct
    ``g2`` with the keyword ``transparent`` set to ``True``::

        sage: g2 = circle((0,0), 1, color='red', transparent=True)
        sage: G = multi_graphics([g1, (g2, (0.2, 0.55, 0.3, 0.3))])
        sage: G
        Multigraphics with 2 elements

    .. PLOT::

        g1 = plot(sin(x**3), (x, -pi, pi))
        g2 = circle((0,0), 1, color='red', transparent=True)
        G = multi_graphics([g1, (g2, (0.2, 0.55, 0.3, 0.3))])
        sphinx_plot(G)

    We can add a new graphics object to G via the method :meth:`append`::

        sage: g3 = complex_plot(zeta, (-20, 10), (-20, 20),
        ....:                   axes_labels=['$x$', '$y$'], frame=True)
        sage: G.append(g3, pos=(0.63, 0.12, 0.3, 0.3))
        sage: G
        Multigraphics with 3 elements

    .. PLOT::

        g1 = plot(sin(x**3), (x, -pi, pi))
        g2 = circle((0,0), 1, color='red', transparent=True)
        G = multi_graphics([g1, (g2, (0.2, 0.55, 0.3, 0.3))])
        g3 = complex_plot(zeta, (-20, 10), (-20, 20), \
                          axes_labels=['$x$', '$y$'], frame=True)
        G.append(g3, pos=(0.63, 0.12, 0.3, 0.3))
        sphinx_plot(G)

    We can access the individual elements composing ``G`` with the
    square-bracket operator::

        sage: print(G[0])
        Graphics object consisting of 1 graphics primitive
        sage: G[0] is g1
        True
        sage: G[1] is g2
        True
        sage: G[2] is g3
        True

    ``G[:]`` returns the full list of graphics objects composing ``G``::

        sage: G[:]
        [Graphics object consisting of 1 graphics primitive,
         Graphics object consisting of 1 graphics primitive,
         Graphics object consisting of 1 graphics primitive]
        sage: len(G)
        3

    """
    def __init__(self, graphics_list):
        r"""
        Initialize the attributes common to all MultiGraphics objects.

        TESTS::

            sage: from sage.plot.multigraphics import MultiGraphics
            sage: G = MultiGraphics([])
            sage: print(G)
            Multigraphics with 0 element
            sage: c = circle((0,0), 1)
            sage: G = MultiGraphics([c, (c, (0.7, 0.6, 0.2, 0.2))])
            sage: print(G)
            Multigraphics with 2 elements

        """
        self._glist = []
        self._positions = []
        #
        for ins in graphics_list:
            if isinstance(ins, Graphics):
                self.append(ins)  # default position
            else:
                if not isinstance(ins, (list, tuple)) or len(ins) != 2:
                    raise TypeError("a pair (Graphics, position) is "
                                    "expected, not {}".format(ins))
                self.append(ins[0], pos=ins[1])

    def _repr_(self):
        r"""
        Representation of ``self``.

        EXAMPLES::

            sage: c = circle((0,0), 1)
            sage: G = graphics_array([c, c, c])
            sage: G._repr_()
            'Graphics Array of size 1 x 3'
            sage: G
            Graphics Array of size 1 x 3

        """
        return str(self)

    def _rich_repr_(self, display_manager, **kwds):
        r"""
        Rich Output Magic Method.

        See :mod:`sage.repl.rich_output` for details.

        .. TODO::

           This method is identical to Graphics._rich_repr_ so it could be
           inherited from a common base class

        EXAMPLES::

            sage: from sage.repl.rich_output import get_display_manager
            sage: dm = get_display_manager()
            sage: G = graphics_array([Graphics(), Graphics()], 1, 2)
            sage: G._rich_repr_(dm)
            OutputImagePng container

        """
        types = display_manager.types
        prefer_raster = (
            ('.png', types.OutputImagePng),
            ('.jpg', types.OutputImageJpg),
            ('.gif', types.OutputImageGif),
        )
        prefer_vector = (
            ('.svg', types.OutputImageSvg),
            ('.pdf', types.OutputImagePdf),
        )
        graphics = display_manager.preferences.graphics
        if graphics == 'disable':
            return
        elif graphics == 'raster' or graphics is None:
            preferred = prefer_raster + prefer_vector
        elif graphics == 'vector':
            preferred = prefer_vector + prefer_raster
        else:
            raise ValueError('unknown graphics output preference')
        for file_ext, output_container in preferred:
            if output_container in display_manager.supported_output():
                return display_manager.graphics_from_save(
                    self.save, kwds, file_ext, output_container)

    def __getitem__(self, i):
        r"""
        Return the ``i``th element of the list of graphics composing ``self``.

        EXAMPLES:

        We can access and view individual plots::

            sage: L = [[plot(x^2)], [plot(x^3)]]
            sage: G = graphics_array(L)
            sage: G[1]
            Graphics object consisting of 1 graphics primitive

        Another example::

            sage: L = [plot(sin(k*x), (x,-pi,pi)) + circle((k,k), 1,
            ....:           color='red') for k in range(10)]
            sage: G = graphics_array(L, 5, 2)
            sage: G[3]
            Graphics object consisting of 2 graphics primitives

        """
        return self._glist[i]

    def __setitem__(self, i, g):
        r"""
        Set the ``i``th element of the list of graphics composing ``self``.

        EXAMPLES::

            sage: L = [[plot(x^2)], [plot(x^3)]]
            sage: G = graphics_array(L)
            sage: G[1] # the plot of x^3
            Graphics object consisting of 1 graphics primitive

        Now we change it::

            sage: G[1] = circle((1,1), 2) + points([(1,2), (3,2), (5,5)],
            ....:                                  color='purple')
            sage: G[1] # a circle and some purple points
            Graphics object consisting of 2 graphics primitives

        """
        self._glist[i] = g

    def __len__(self):
        r"""
        Total number of Graphics objects composing ``self``.

        EXAMPLES::

            sage: L = [circle((0,0), n) for n in range(6)]
            sage: G = graphics_array(L, 2, 3)
            sage: len(G)
            6

        """
        return len(self._glist)

    def matplotlib(self, figure=None, figsize=None, **kwds):
        r"""
        Construct or modify a Matplotlib figure by drawing ``self`` on it.

        INPUT:

        - ``figure`` -- (default: ``None``) Matplotlib figure (class
          ``matplotlib.figure.Figure``) on which ``self`` is to be displayed;
          if ``None``, the figure will be created from the parameter
          ``figsize``

        - ``figsize`` -- (default: ``None``) width or [width, height] in inches
          of the Matplotlib figure in case ``figure`` is ``None``; if
          ``figsize`` is ``None``, Matplotlib's default (6.4 x 4.8 inches) is
          used

        - ``kwds`` -- options passed to the
          :meth:`~sage.plot.graphics.Graphics.matplotlib` method of
          each graphics object constituting ``self``

        OUTPUT:

        - a ``matplotlib.figure.Figure`` object; if the argument ``figure`` is
          provided, this is the same object as ``figure``.

        EXAMPLES:

        Let us consider a :class:`GraphicsArray` object with 3 elements::

            sage: G = graphics_array([plot(sin(x^k), (x, 0, 3))
            ....:                     for k in range(1, 4)])

        If ``matplotlib()`` is invoked without any argument, a Matplotlib
        figure is created and contains the 3 graphics element of the array
        as 3 Matplotlib ``Axes``::

            sage: fig = G.matplotlib()
            sage: fig
            <Figure size 640x480 with 3 Axes>
            sage: type(fig)
            <class 'matplotlib.figure.Figure'>

        Specifying the figure size (in inches)::

            sage: G.matplotlib(figsize=(8., 5.))
            <Figure size 800x500 with 3 Axes>

        If a single number is provided for ``figsize``, it is considered to be
        the width; the height is then computed according to Matplotlib's
        default aspect ratio (4/3)::

            sage: G.matplotlib(figsize=8.)
            <Figure size 800x600 with 3 Axes>

        An example of use with a preexisting created figure, created by
        ``pyplot``::

            sage: import matplotlib.pyplot as plt
            sage: fig1 = plt.figure(1)
            sage: fig1
            <Figure size 640x480 with 0 Axes>
            sage: fig_out = G.matplotlib(figure=fig1)
            sage: fig_out
            <Figure size 640x480 with 3 Axes>

        Note that the output figure is the same object as the input one::

            sage: fig_out is fig1
            True

        It has however been modified by ``G.matplotlib(figure=fig1)``, which
        has added 3 new ``Axes`` to it.

        Another example, with a figure created from scratch, via Matplolib's
        ``Figure``::

            sage: from matplotlib.figure import Figure
            sage: fig2 = Figure()
            sage: fig2
            <Figure size 640x480 with 0 Axes>
            sage: G.matplotlib(figure=fig2)
            <Figure size 640x480 with 3 Axes>
            sage: fig2
            <Figure size 640x480 with 3 Axes>

        """
        from matplotlib.figure import Figure
        glist = self._glist
        if len(glist) == 0:       # for an empty MultiGraphics, we create
            glist = [Graphics()]  # a 1-element list with an empty graphics
        # If no Matplotlib figure is provided, it is created here:
        if figure is None:
            if figsize is not None:
                figsize = _parse_figsize(figsize)
            figure = Figure(figsize=figsize)
        global do_verify
        do_verify = True
        for i, g in enumerate(glist):
            # Options for g.matplotlib():
            options = {}
            options.update(Graphics.SHOW_OPTIONS)  # default options for show()
            options['legend_options'] = Graphics.LEGEND_OPTIONS  # default leg.
            options.update(g._extra_kwds)  # options set in g
            options.update(kwds)
            # We get rid of options that are not relevant for g.matplotlib():
            options.pop('dpi', None)
            options.pop('fig_tight', None)
            transparent = options.pop('transparent', None)
            # Creating the Matplotlib Axes object "subplot" on the figure:
            subplot = self._add_subplot(figure, i)
            # and drawing g on it:
            g.matplotlib(figure=figure, sub=subplot, verify=do_verify,
                         **options)
            if transparent:
                subplot.set_facecolor('none')
        return figure

    def save(self, filename, figsize=None, **kwds):
        r"""
        Save ``self`` to a file, in various formats.

        INPUT:

        - ``filename`` -- (string) the file name; the image format is given by
          the extension, which can be one of the following:

            * ``.eps``,

            * ``.pdf``,

            * ``.png``,

            * ``.ps``,

            * ``.sobj`` (for a Sage object you can load later),

            * ``.svg``,

            * empty extension will be treated as ``.sobj``.

        - ``figsize`` -- (default: ``None``) width or [width, height] in inches
          of the Matplotlib figure; if none is provided, Matplotlib's default
          (6.4 x 4.8 inches) is used

        - ``kwds`` -- keyword arguments, like ``dpi=...``, passed to the
          plotter, see :meth:`show`

        EXAMPLES::

            sage: F = tmp_filename(ext='.png')
            sage: L = [plot(sin(k*x), (x,-pi,pi)) for k in [1..3]]
            sage: G = graphics_array(L)
            sage: G.save(F, dpi=500, axes=False)

        TESTS::

            sage: graphics_array([]).save(F)
            sage: graphics_array([[]]).save(F)

        """
        from matplotlib import rcParams
        ext = os.path.splitext(filename)[1].lower()
        if ext in ['', '.sobj']:
            SageObject.save(self, filename)
        elif ext not in ALLOWED_EXTENSIONS:
            raise ValueError("allowed file extensions for images are '" +
                             "', '".join(ALLOWED_EXTENSIONS) + "'!")
        else:
            rc_backup = (rcParams['ps.useafm'], rcParams['pdf.use14corefonts'],
                         rcParams['text.usetex'])  # save the rcParams
            figure = self.matplotlib(figsize=figsize, **kwds)
            transparent = kwds.get('transparent',
                                   Graphics.SHOW_OPTIONS['transparent'])
            fig_tight = kwds.get('fig_tight',
                                 Graphics.SHOW_OPTIONS['fig_tight'])
            dpi = kwds.get('dpi', Graphics.SHOW_OPTIONS['dpi'])
            # One can output in PNG, PS, EPS, PDF, PGF, or SVG format,
            # depending on the file extension.
            # PGF is handled by a different backend
            if ext == '.pgf':
                from sage.misc.sage_ostools import have_program
                latex_implementations = [i for i in ["xelatex", "pdflatex",
                                                     "lualatex"]
                                         if have_program(i)]
                if not latex_implementations:
                    raise ValueError("Matplotlib requires either xelatex, "
                                     "lualatex, or pdflatex.")
                if latex_implementations[0] == "pdflatex":
                    # use pdflatex and set font encoding as per
                    # Matplotlib documentation:
                    # https://matplotlib.org/users/pgf.html#pgf-tutorial
                    pgf_options = {"pgf.texsystem": "pdflatex",
                                   "pgf.preamble": [
                                      r"\usepackage[utf8x]{inputenc}",
                                      r"\usepackage[T1]{fontenc}"
                                      ]}
                else:
                    pgf_options = {"pgf.texsystem": latex_implementations[0]}
                rcParams.update(pgf_options)
                from matplotlib.backends.backend_pgf import FigureCanvasPgf
                figure.set_canvas(FigureCanvasPgf(figure))
            # Matplotlib looks at the file extension to see what the renderer
            # should be. The default is FigureCanvasAgg for PNG's because this
            # is by far the most common type of files rendered, like in the
            # notebook, for example. If the file extension is not '.png', then
            # Matplotlib will handle it.
            else:
                from matplotlib.backends.backend_agg import FigureCanvasAgg
                figure.set_canvas(FigureCanvasAgg(figure))
            if isinstance(self, GraphicsArray):
                # tight_layout adjusts the *subplot* parameters so ticks aren't
                # cut off, etc.
                figure.tight_layout()
            opts = dict(dpi=dpi, transparent=transparent)
            if fig_tight is True:
                opts['bbox_inches'] = 'tight'
            figure.savefig(filename, **opts)
            # Restore the rcParams to the original, possibly user-set values
            (rcParams['ps.useafm'], rcParams['pdf.use14corefonts'],
             rcParams['text.usetex']) = rc_backup

    def save_image(self, filename=None, *args, **kwds):
        r"""
        Save an image representation of ``self``.  The image type is
        determined by the extension of the filename.  For example,
        this could be ``.png``, ``.jpg``, ``.gif``, ``.pdf``,
        ``.svg``.  Currently this is implemented by calling the
        :meth:`save` method of self, passing along all arguments and
        keywords.

        .. NOTE::

            Not all image types are necessarily implemented for all
            graphics types.  See :meth:`save` for more details.

        EXAMPLES::

            sage: plots = [[plot(m*cos(x + n*pi/4), (x, 0, 2*pi))
            ....:           for n in range(3)] for m in range(1,3)]
            sage: G = graphics_array(plots)
            sage: G.save_image(tmp_filename(ext='.png'))

        """
        self.save(filename, *args, **kwds)

    def _latex_(self, **kwds):
        r"""
        Return a string plotting ``self`` with PGF.

        INPUT:

        All keyword arguments will be passed to the plotter.

        OUTPUT:

        A string of PGF commands to plot ``self``

        EXAMPLES::

            sage: A = graphics_array([plot(sin), plot(cos)])
            sage: A._latex_()[:40]  # not tested (see comment below)
            '%% Creator: Matplotlib, PGF backend\n%%\n%'

        The above doctest fails on macOS due to the following Matplotlib issue: https://github.com/matplotlib/matplotlib/issues/10307

        """
        tmpfilename = tmp_filename(ext='.pgf')
        self.save(filename=tmpfilename, **kwds)
        with open(tmpfilename, "r") as tmpfile:
            latex_list = tmpfile.readlines()
        return ''.join(latex_list)

    def show(self, **kwds):
        r"""
        Show ``self`` immediately.

        This method attempts to display the graphics immediately,
        without waiting for the currently running code (if any) to
        return to the command line. Be careful, calling it from within
        a loop will potentially launch a large number of external
        viewer programs.

        OPTIONAL INPUT:

        - ``dpi`` -- dots per inch

        - ``figsize`` -- width or [width, height] of the figure, in inches; the
          default is 6.4 x 4.8 inches

        - ``axes`` -- boolean; if ``True``, all individual graphics are
          endowed with axes; if ``False``, all axes are removed (this overrides
          the ``axes`` option set in each graphics)

        - ``frame`` -- boolean; if ``True``, all individual graphics are
          drawn with a frame around them; if ``False``, all frames are removed
          (this overrides the ``frame`` option set in each graphics)

        - ``fontsize`` -- positive integer, the size of fonts for the axes
          labels (this overrides the ``fontsize`` option set in each graphics)

        OUTPUT:

        This method does not return anything. Use :meth:`save` if you
        want to save the figure as an image.

        EXAMPLES:

        This draws a graphics array with four trig plots and no axes in any of
        the plots and a figure width of 4 inches::

            sage: G = graphics_array([[plot(sin), plot(cos)],
            ....:                     [plot(tan), plot(sec)]])
            sage: G.show(axes=False, figsize=4)

        .. PLOT::

            G = graphics_array([[plot(sin), plot(cos)], \
                                [plot(tan), plot(sec)]])
            sphinx_plot(G, axes=False, figsize=4)

        Same thing with a frame around each individual graphics::

            sage: G.show(axes=False, frame=True, figsize=4)

        .. PLOT::

            G = graphics_array([[plot(sin), plot(cos)], \
                                [plot(tan), plot(sec)]])
            sphinx_plot(G, axes=False, frame=True, figsize=4)

        Actually, many options are possible; for instance, we may set
        ``fontsize`` and ``gridlines``::

            sage: G.show(axes=False, frame=True, figsize=4, fontsize=8,
            ....:        gridlines='major')

        .. PLOT::

            G = graphics_array([[plot(sin), plot(cos)], \
                                [plot(tan), plot(sec)]])
            sphinx_plot(G, axes=False, frame=True, figsize=4, fontsize=8, \
                        gridlines='major')

        """
        from sage.repl.rich_output import get_display_manager
        dm = get_display_manager()
        dm.display_immediately(self, **kwds)

    def plot(self):
        r"""
        Return ``self`` since ``self`` is already a graphics object.

        EXAMPLES::

            sage: g1 = plot(cos, 0, 1)
            sage: g2 = circle((0,0), 1)
            sage: G = multi_graphics([g1, g2])
            sage: G.plot() is G
            True

        """
        return self

    def inset(self, graphics, pos=None, fontsize=None):
        r"""
        Add a graphics object as an inset.

        INPUT:

        - ``graphics`` -- the graphics object (instance of :class:`Graphics`)
          to be added as an inset

        - ``pos`` -- (default: ``None``) 4-tuple
          ``(left, bottom, width, height)`` specifying the location and
          relative size of the inset on the canvas, all quantities being
          expressed in fractions of the canvas width and height; if ``None``,
          the value ``(0.7, 0.7, 0.2, 0.2)`` is used

        - ``fontsize`` -- (default: ``None``)  integer, font size (in points)
          for the inset; if ``None``, the value of 6 points is used, unless
          ``fontsize`` has been explicitly set in the construction of
          ``graphics`` (in this case, it is not overwritten here)

        OUTPUT:

        - instance of :class:`~sage.plot.multigraphics.MultiGraphics`

        EXAMPLES:

        Let us consider a graphics array of 2 elements::

            sage: G = graphics_array([plot(sin, (0, 2*pi)),
            ....:                     plot(cos, (0, 2*pi))])
            sage: G
            Graphics Array of size 1 x 2

        .. PLOT::

            G = graphics_array([plot(sin, (0, 2*pi)), plot(cos, (0, 2*pi))])
            sphinx_plot(G)

        and add some inset at the default position::

            sage: c = circle((0,0), 1, color='red', thickness=2, frame=True)
            sage: G.inset(c)
            Multigraphics with 3 elements

        .. PLOT::

            G = graphics_array([plot(sin, (0, 2*pi)), plot(cos, (0, 2*pi))])
            c = circle((0,0), 1, color='red', thickness=2, frame=True)
            sphinx_plot(G.inset(c))

        We may customize the position and font size of the inset::

            sage: G.inset(c, pos=(0.3, 0.7, 0.2, 0.2), fontsize=8)
            Multigraphics with 3 elements

        .. PLOT::

            G = graphics_array([plot(sin, (0, 2*pi)), plot(cos, (0, 2*pi))])
            c = circle((0,0), 1, color='red', thickness=2, frame=True)
            sphinx_plot(G.inset(c, pos=(0.3, 0.7, 0.2, 0.2), fontsize=8))

        """
        if pos is None:
            pos = (0.7, 0.7, 0.2, 0.2)
        if fontsize is not None:
            graphics._extra_kwds['fontsize'] = fontsize
        elif 'fontsize' not in graphics._extra_kwds:
            graphics._extra_kwds['fontsize'] = 6
        current = []  # list of current pairs (graphics, position)
        for i, g in enumerate(self._glist):
            current.append((g, self.position(i)))
        resu = MultiGraphics(current)
        resu.append(graphics, pos=pos)
        return resu

    #
    # Methods to reimplement in derived classes:
    #
    def __str__(self):
        r"""
        String representation of ``self``

        EXAMPLES::

            sage: from sage.plot.multigraphics import MultiGraphics
            sage: G = MultiGraphics([])
            sage: G.__str__()
            'Multigraphics with 0 element'
            sage: str(G)
            'Multigraphics with 0 element'
            sage: c = circle((0,0), 1)
            sage: G = MultiGraphics([c])
            sage: str(G)
            'Multigraphics with 1 element'
            sage: G = MultiGraphics([c, c, c])
            sage: str(G)
            'Multigraphics with 3 elements'

        """
        n = len(self._glist)
        if n <= 1:
            return "Multigraphics with {} element".format(n)
        return "Multigraphics with {} elements".format(n)

    def _add_subplot(self, figure, index, **options):
        r"""
        Add a subplot to a given Matplotlib ``Figure``, the position of
        which is governed by a given element of ``self``.

        This method encapsulates the Matplotlib method ``Figure.add_axes``
        and is intended to be called by :meth:`MultiGraphics.save`.

        INPUT:

        - ``figure`` -- a Matplotlib ``Figure`` object
        - ``index`` -- integer specifiying the element of ``self``
        - ``options`` -- extra options to be passed to ``Figure.add_axes``

        OUTPUT:

        - a Matplotlib ``Axes`` object

        EXAMPLES::

            sage: g0 = circle((0,0), 1)
            sage: g1 = plot(sin)
            sage: G = multi_graphics([g0, (g1, (0.2, 0.3, 0.4, 0.1))])
            sage: from matplotlib.figure import Figure
            sage: fig = Figure()
            sage: fig
            <Figure size 640x480 with 0 Axes>
            sage: ax0 = G._add_subplot(fig, 0)
            sage: type(ax0)
            <class 'matplotlib.axes._axes.Axes'>
            sage: fig
            <Figure size 640x480 with 1 Axes>
            sage: ax1 = G._add_subplot(fig, 1)
            sage: fig
            <Figure size 640x480 with 2 Axes>

        TESTS::

            sage: [ax0, ax1] == fig.get_axes()
            True
            sage: G.position(1)
            (0.2, 0.3, 0.4, 0.1)
            sage: ax1.get_position().bounds  # tol 1.0e-13
            (0.2, 0.3, 0.4000000000000001, 0.10000000000000003)

        """
        # Note: using label=str(index) ensures that a new Axes is generated
        # for each element of ``self``, even if some elements share the same
        # positions
        return figure.add_axes(self._positions[index], label=str(index),
                               **options)

    def position(self, index):
        r"""
        Return the position and relative size of an element of ``self`` on the
        canvas.

        INPUT:

        - ``index`` -- integer specifiying which element of ``self``

        OUTPUT:

        - a 4-tuple ``(left, bottom, width, height)`` giving the location and
          relative size of the element on the canvas, all quantities being
          expressed in fractions of the canvas width and height

        EXAMPLES::

            sage: g1 = plot(sin(x^2), (x, 0, 4))
            sage: g2 = circle((0,0), 1, rgbcolor='red', fill=True, axes=False)
            sage: G = multi_graphics([g1, (g2, (0.15, 0.2, 0.1, 0.15))])
            sage: G.position(0)  # tol 1.0e-13
            (0.125, 0.11, 0.775, 0.77)
            sage: G.position(1)  # tol 1.0e-13
            (0.15, 0.2, 0.1, 0.15)

        """
        return self._positions[index]

    def append(self, graphics, pos=None):
        r"""
        Append a graphics object to ``self``.

        INPUT:

        - ``graphics`` -- the graphics object (instance of :class:`Graphics`)
          to be added to ``self``

        - ``pos`` -- (default: ``None``) 4-tuple
          ``(left, bottom, width, height)`` specifying the location and size
          of ``graphics`` on the canvas, all quantities being in fractions of
          the canvas width and height; if ``None``, ``graphics`` is assumed to
          occupy the whole canvas, except for some padding; this corresponds to
          the default position
          ``(left, bottom, width, height) = (0.125, 0.11, 0.775, 0.77)``

        EXAMPLES:

        Let us consider a multigraphics with 2 elements::

            sage: g1 = plot(chebyshev_T(4, x), (x, -1, 1), title='n=4')
            sage: g2 = plot(chebyshev_T(8, x), (x, -1, 1), title='n=8',
            ....:           color='red')
            sage: G = multi_graphics([(g1, (0.125, 0.2, 0.4, 0.4)),
            ....:                     (g2, (0.55, 0.4, 0.4, 0.4))])
            sage: G
            Multigraphics with 2 elements

        .. PLOT::

            g1 = plot(chebyshev_T(4, x), (x, -1, 1), title='n=4')
            g2 = plot(chebyshev_T(8, x), (x, -1, 1), title='n=8', color='red')
            G = multi_graphics([(g1, (0.125, 0.2, 0.4, 0.4)), \
                                (g2, (0.55, 0.4, 0.4, 0.4))])
            sphinx_plot(G)

        We append a third plot to it::

            sage: g3 = plot(chebyshev_T(16, x), (x, -1, 1), title='n=16',
            ....:           color='brown')
            sage: G.append(g3, pos=(0.55, 0.11, 0.4, 0.15))
            sage: G
            Multigraphics with 3 elements

        .. PLOT::

            g1 = plot(chebyshev_T(4, x), (x, -1, 1), title='n=4')
            g2 = plot(chebyshev_T(8, x), (x, -1, 1), title='n=8', color='red')
            G = multi_graphics([(g1, (0.125, 0.2, 0.4, 0.4)), \
                                (g2, (0.55, 0.4, 0.4, 0.4))])
            g3 = plot(chebyshev_T(16, x), (x, -1, 1), title='n=16', \
                      color='brown')
            G.append(g3, pos=(0.55, 0.11, 0.4, 0.15))
            sphinx_plot(G)

        We may use ``append`` to add a title::

            sage: title = text("Chebyshev polynomials", (0, 0), fontsize=16,
            ....:              axes=False)
            sage: G.append(title, pos=(0.18, 0.8, 0.7, 0.1))
            sage: G
            Multigraphics with 4 elements

        .. PLOT::

            g1 = plot(chebyshev_T(4, x), (x, -1, 1), title='n=4')
            g2 = plot(chebyshev_T(8, x), (x, -1, 1), title='n=8', color='red')
            G = multi_graphics([(g1, (0.125, 0.2, 0.4, 0.4)), \
                                (g2, (0.55, 0.4, 0.4, 0.4))])
            g3 = plot(chebyshev_T(16, x), (x, -1, 1), title='n=16', \
                      color='brown')
            G.append(g3, pos=(0.55, 0.11, 0.4, 0.15))
            title = text("Chebyshev polynomials", (0, 0), fontsize=16, \
                         axes=False)
            G.append(title, pos=(0.18, 0.8, 0.7, 0.1))
            sphinx_plot(G)

        .. SEEALSO::

            :meth:`inset`

        """
        from matplotlib import rcParams
        if not isinstance(graphics, Graphics):
            raise TypeError("a Graphics object is expected, "
                            "not {}".format(graphics))
        if pos is None:
            # Default position:
            left = rcParams['figure.subplot.left']
            bottom = rcParams['figure.subplot.bottom']
            width = rcParams['figure.subplot.right'] - left
            height = rcParams['figure.subplot.top'] - bottom
            pos = (left, bottom, width, height)
        elif not isinstance(pos, (list, tuple)) or len(pos) != 4:
            raise TypeError("pos must be a 4-tuple, not {}".format(pos))
        pos = tuple(float(p) for p in pos)
        self._glist.append(graphics)
        self._positions.append(pos)


# ****************************************************************************


class GraphicsArray(MultiGraphics):
    r"""
    This class implements 2-dimensional graphical objects that constitute
    an array of :class:`~sage.plot.graphics.Graphics` drawn on a single
    canvas.

    The user interface is through the function
    :func:`~sage.plot.plot.graphics_array`.

    INPUT:

    - ``array`` -- either a list of lists of
      :class:`~sage.plot.graphics.Graphics` elements (generic case) or a
      single list of :class:`~sage.plot.graphics.Graphics` elements (case of a
      single-row array)

    EXAMPLES:

    An array made of four graphics objects::

        sage: g1 = plot(sin(x^2), (x, 0, 6), axes_labels=['$x$', '$y$'],
        ....:           axes=False, frame=True, gridlines='minor')
        sage: y = var('y')
        sage: g2 = streamline_plot((sin(x), cos(y)), (x,-3,3), (y,-3,3),
        ....:                      aspect_ratio=1)
        sage: g3 = graphs.DodecahedralGraph().plot()
        sage: g4 = polar_plot(sin(5*x)^2, (x, 0, 2*pi), color='green',
        ....:                 fontsize=8) \
        ....:      + circle((0,0), 0.5, rgbcolor='red', fill=True, alpha=0.1,
        ....:               legend_label='pink')
        sage: g4.set_legend_options(loc='upper right')
        sage: G = graphics_array([[g1, g2], [g3, g4]])
        sage: G
        Graphics Array of size 2 x 2

    .. PLOT::

        g1 = plot(sin(x**2), (x, 0, 6), axes_labels=['$x$', '$y$'], \
                  axes=False, frame=True, gridlines='minor')
        y = var('y')
        g2 = streamline_plot((sin(x), cos(y)), (x,-3,3), (y,-3,3), \
                             aspect_ratio=1)
        g3 = graphs.DodecahedralGraph().plot()
        g4 = polar_plot(sin(5*x)**2, (x, 0, 2*pi), color='green', fontsize=8) \
             + circle((0,0), 0.5, rgbcolor='red', fill=True, alpha=0.1, \
                      legend_label='pink')
        g4.set_legend_options(loc='upper right')
        G = graphics_array([[g1, g2], [g3, g4]])
        sphinx_plot(G)

    If one constructs the graphics array from a single list of graphics
    objects, one obtains a single-row array::

        sage: G = graphics_array([g1, g2, g3, g4])
        sage: G
        Graphics Array of size 1 x 4

    .. PLOT::

        g1 = plot(sin(x**2), (x, 0, 6), axes_labels=['$x$', '$y$'], \
                  axes=False, frame=True, gridlines='minor')
        y = var('y')
        g2 = streamline_plot((sin(x), cos(y)), (x,-3,3), (y,-3,3), \
                             aspect_ratio=1)
        g3 = graphs.DodecahedralGraph().plot()
        g4 = polar_plot(sin(5*x)**2, (x, 0, 2*pi), color='green', fontsize=8) \
             + circle((0,0), 0.5, rgbcolor='red', fill=True, alpha=0.1, \
                      legend_label='pink')
        g4.set_legend_options(loc='upper right')
        G = graphics_array([g1, g2, g3, g4])
        sphinx_plot(G)

    We note that the overall aspect ratio of the figure is 4/3 (the default),
    which makes ``g1`` elongated, while the aspect ratio of ``g2``, which has
    been specified with the parameter ``aspect_ratio=1`` is preserved. To get
    a better aspect ratio for the whole figure, one can use the option
    ``figsize`` in the method :meth:`~MultiGraphics.show`::

        sage: G.show(figsize=[8, 3])

    .. PLOT::

        g1 = plot(sin(x**2), (x, 0, 6), axes_labels=['$x$', '$y$'], \
                  axes=False, frame=True, gridlines='minor')
        y = var('y')
        g2 = streamline_plot((sin(x), cos(y)), (x,-3,3), (y,-3,3), \
                             aspect_ratio=1)
        g3 = graphs.DodecahedralGraph().plot()
        g4 = polar_plot(sin(5*x)**2, (x, 0, 2*pi), color='green', fontsize=8) \
             + circle((0,0), 0.5, rgbcolor='red', fill=True, alpha=0.1, \
                      legend_label='pink')
        g4.set_legend_options(loc='upper right')
        G = graphics_array([g1, g2, g3, g4])
        sphinx_plot(G, figsize=[8, 3])

    We can access individual elements of the graphics array with the
    square-bracket operator::

        sage: G = graphics_array([[g1, g2], [g3, g4]])  # back to the 2x2 array
        sage: print(G)
        Graphics Array of size 2 x 2
        sage: G[0] is g1
        True
        sage: G[1] is g2
        True
        sage: G[2] is g3
        True
        sage: G[3] is g4
        True

    Note that with respect to the square-bracket operator, ``G`` is considered
    as a flattened list of graphics objects, not as an array. For instance,
    ``G[0, 1]`` throws an error::

        sage: G[0, 1]  # py3 (error message is slightly different with Python 2)
        Traceback (most recent call last):
        ...
        TypeError: list indices must be integers or slices, not tuple

    ``G[:]`` returns the full (flattened) list of graphics objects composing
    ``G``::

        sage: G[:]
        [Graphics object consisting of 1 graphics primitive,
        Graphics object consisting of 1 graphics primitive,
        Graphics object consisting of 51 graphics primitives,
        Graphics object consisting of 2 graphics primitives]

    The total number of Graphics objects composing the array is returned
    by the function ``len``::

        sage: len(G)
        4

    The square-bracket operator can be used to replace elements in the array::

        sage: G[0] = g4
        sage: G
        Graphics Array of size 2 x 2

    .. PLOT::

        g1 = plot(sin(x**2), (x, 0, 6), axes_labels=['$x$', '$y$'], \
                  axes=False, frame=True, gridlines='minor')
        y = var('y')
        g2 = streamline_plot((sin(x), cos(y)), (x,-3,3), (y,-3,3), \
                             aspect_ratio=1)
        g3 = graphs.DodecahedralGraph().plot()
        g4 = polar_plot(sin(5*x)**2, (x, 0, 2*pi), color='green', fontsize=8) \
             + circle((0,0), 0.5, rgbcolor='red', fill=True, alpha=0.1, \
                      legend_label='pink')
        g4.set_legend_options(loc='upper right')
        G = graphics_array([[g1, g2], [g3, g4]])
        G[0] = g4
        sphinx_plot(G)

    """
    def __init__(self, array):
        r"""
        Construct a ``GraphicsArray``.

        TESTS::

            sage: from sage.plot.multigraphics import GraphicsArray
            sage: g = circle((0,0), 1)  # a Graphics object
            sage: G = GraphicsArray([[g, g], [g, g]])
            sage: print(G)
            Graphics Array of size 2 x 2

        Construction from a single list ==> 1-row array::

            sage: G = GraphicsArray([g, g, g])
            sage: print(G)
            Graphics Array of size 1 x 3
            sage: G = GraphicsArray([g])
            sage: print(G)
            Graphics Array of size 1 x 1

        Empty array::

            sage: G = GraphicsArray([])
            sage: print(G)
            Graphics Array of size 0 x 0
            sage: len(G)
            0
            sage: G = GraphicsArray([[]])
            sage: print(G)
            Graphics Array of size 1 x 0
            sage: len(G)
            0

        Check treatment of wrong inputs::

            sage: G = GraphicsArray([[g, g], [g]])
            Traceback (most recent call last):
            ...
            TypeError: array must be a list of equal-size lists of Graphics
             objects, not [[Graphics object consisting of 1 graphics primitive,
             Graphics object consisting of 1 graphics primitive],
             [Graphics object consisting of 1 graphics primitive]]
            sage: G = GraphicsArray(g)
            Traceback (most recent call last):
            ...
            TypeError: array must be a list of lists of Graphics objects, not
             Graphics object consisting of 1 graphics primitive
            sage: G = GraphicsArray([g, x])
            Traceback (most recent call last):
            ...
            TypeError: every element of array must be a Graphics object

        """
        MultiGraphics.__init__(self, [])
        if not isinstance(array, (list, tuple)):
            raise TypeError("array must be a list of lists of Graphics "
                            "objects, not {}".format(array))
        array = list(array)
        self._rows = len(array)
        if self._rows > 0:
            if not isinstance(array[0], (list, tuple)):
                array = [array]
                self._rows = 1
            self._cols = len(array[0])
        else:
            self._cols = 0
        for row in array:  # basically flatten the list
            if not isinstance(row, (list, tuple)) or len(row) != self._cols:
                raise TypeError("array must be a list of equal-size lists of "
                                "Graphics objects, not {}".format(array))
            for g in row:
                if not isinstance(g, Graphics):
                    raise TypeError("every element of array must be a "
                                    "Graphics object")
                self._glist.append(g)
        # self._positions is not initialized since most of the time, it is not
        # not used. It is required only by the method inset(); it is then
        # initialized by the method position().

    def __str__(self):
        r"""
        String representation of the graphics array.

        EXAMPLES::

            sage: c = circle((0,0), 1)
            sage: G = graphics_array([c]*6, 2, 3)
            sage: G.__str__()
            'Graphics Array of size 2 x 3'
            sage: str(G)
            'Graphics Array of size 2 x 3'

        """
        return "Graphics Array of size {} x {}".format(self._rows, self._cols)

    def _add_subplot(self, figure, index, **options):
        r"""
        Add a subplot to a given Matplotlib ``Figure``, the position of
        which is governed by a given element of ``self``.

        This method encapsulates the Matplotlib method ``Figure.add_subplot``
        and is intended to be called by :meth:`MultiGraphics.save`.

        INPUT:

        - ``figure`` -- a Matplotlib ``Figure`` object
        - ``index`` -- integer specifiying the element of ``self``
        - ``options`` -- extra options to be passed to ``Figure.add_subplot``

        OUTPUT:

        - a Matplotlib ``Axes`` object

        EXAMPLES::

            sage: c = circle((0,0), 1)
            sage: G = graphics_array([c, c])
            sage: from matplotlib.figure import Figure
            sage: fig = Figure()
            sage: ax1 = G._add_subplot(fig, 0)
            sage: type(ax1)
            <class 'matplotlib.axes._subplots.AxesSubplot'>
            sage: ax2 = G._add_subplot(fig, 1)
            sage: fig.get_axes() == [ax1, ax2]
            True

        """
        if self._rows == 0 or self._cols == 0:
            rows = 1
            cols = 1
        else:
            rows = self._rows
            cols = self._cols
        # index --> index + 1 for Figure.add_subplot:
        return figure.add_subplot(rows, cols, index + 1, **options)

    def nrows(self):
        r"""
        Number of rows of the graphics array.

        EXAMPLES::

            sage: R = rainbow(6)
            sage: L = [plot(x^n, (x,0,1), color=R[n]) for n in range(6)]
            sage: G = graphics_array(L, 2, 3)
            sage: G.nrows()
            2
            sage: graphics_array(L).nrows()
            1

        """
        return self._rows

    def ncols(self):
        r"""
        Number of columns of the graphics array.

        EXAMPLES::

            sage: R = rainbow(6)
            sage: L = [plot(x^n, (x,0,1), color=R[n]) for n in range(6)]
            sage: G = graphics_array(L, 2, 3)
            sage: G.ncols()
            3
            sage: graphics_array(L).ncols()
            6
        """
        return self._cols

    def append(self, g):
        r"""
        Append a graphics to the array.

        Currently not implemented.

        TESTS::

            sage: from sage.plot.multigraphics import GraphicsArray
            sage: c = circle((0,0), 1)
            sage: G = GraphicsArray([c, c])
            sage: G.append(c)
            Traceback (most recent call last):
            ...
            NotImplementedError: Appending to a graphics array is not yet
             implemented

        """
        # Not clear if there is a way to do this
        raise NotImplementedError('Appending to a graphics array is not '
                                  'yet implemented')

    def position(self, index):
        r"""
        Return the position and relative size of an element of ``self`` on the
        canvas.

        INPUT:

        - ``index`` -- integer specifiying which element of ``self``

        OUTPUT:

        - a 4-tuple ``(left, bottom, width, height)`` giving the location and
          relative size of the element on the canvas, all quantities being
          expressed in fractions of the canvas width and height

        EXAMPLES::

            sage: g1 = plot(sin(x), (x, -pi, pi))
            sage: g2 = circle((0,1), 1.)
            sage: G = graphics_array([g1, g2])
            sage: G.position(0)  # tol 5.0e-3
            (0.025045451349937315,
             0.03415488992713045,
             0.4489880779745068,
             0.9345951100728696)
            sage: G.position(1)  # tol 5.0e-3
            (0.5170637412999687,
             0.20212705964722733,
             0.4489880779745068,
             0.5986507706326758)

        """
        if not self._positions:
            # self._positions must be generated, by invoking get_position() on
            # each of the Axes of the Matplotlib figure corresponding to self:
            from matplotlib.backends.backend_agg import FigureCanvasAgg
            figure = self.matplotlib()
            figure.set_canvas(FigureCanvasAgg(figure))
            figure.tight_layout()
            axes = figure.get_axes()
            self._positions = [ax.get_position().bounds for ax in axes]
        return self._positions[index]



"""
HTML Fragments

This module defines a HTML fragment class, which holds a piece of
HTML. This is primarily used in browser-based notebooks, though it
might be useful for creating static pages as well.
"""
from __future__ import absolute_import

#*****************************************************************************
#       Copyright (C) 2008 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import warnings
from sage.misc.latex import latex
from sage.misc.sage_eval import sage_eval
from sage.structure.sage_object import SageObject


class HtmlFragment(str, SageObject):
    r"""
    A HTML fragment.

    This is a piece of HTML, usually not a complete document.  For
    example, just a ``<div>...</div>`` piece and not the entire
    ``<html>...</html>``.

    EXAMPLES::

        sage: from sage.misc.html import HtmlFragment
        sage: HtmlFragment('<b>test</b>')
        <b>test</b>

    .. automethod:: _rich_repr_
    """

    def _rich_repr_(self, display_manager, **kwds):
        """
        Rich Output Magic Method

        See :mod:`sage.repl.rich_output` for details.

        EXAMPLES::

            sage: from sage.repl.rich_output import get_display_manager
            sage: dm = get_display_manager()
            sage: h = sage.misc.html.HtmlFragment('<b>old</b>')
            sage: h._rich_repr_(dm)    # the doctest backend does not support html
            OutputPlainText container
        """
        OutputHtml = display_manager.types.OutputHtml
        if OutputHtml in display_manager.supported_output():
            return OutputHtml(self)
        else:
            return display_manager.types.OutputPlainText(self)


def math_parse(s):
    r"""
    Replace TeX-``$`` with Mathjax equations.
    
    Turn the HTML-ish string s that can have \$\$ and \$'s in it into
    pure HTML.  See below for a precise definition of what this means.

    INPUT:

    - ``s`` -- a string

    OUTPUT:

    A :class:`HtmlFragment` instance.

    Do the following:

    * Replace all ``\$ text \$``\'s by
      ``<script type="math/tex"> text </script>``
    * Replace all ``\$\$ text \$\$``\'s by
      ``<script type="math/tex; mode=display"> text </script>``
    * Replace all ``\ \$``\'s by ``\$``\'s.  Note that in
      the above two cases nothing is done if the ``\$``
      is preceeded by a backslash.
    * Replace all ``\[ text \]``\'s by
      ``<script type="math/tex; mode=display"> text </script>``

    EXAMPLES::

        sage: pretty_print(sage.misc.html.math_parse('This is $2+2$.'))
        This is <script type="math/tex">2+2</script>.
        sage: pretty_print(sage.misc.html.math_parse('This is $$2+2$$.'))
        This is <script type="math/tex; mode=display">2+2</script>.
        sage: pretty_print(sage.misc.html.math_parse('This is \\[2+2\\].'))
        This is <script type="math/tex; mode=display">2+2</script>.
        sage: pretty_print(sage.misc.html.math_parse(r'This is \[2+2\].'))
        This is <script type="math/tex; mode=display">2+2</script>.

    TESTS::

        sage: sage.misc.html.math_parse(r'This \$\$is $2+2$.')
        This $$is <script type="math/tex">2+2</script>.
    """
    # first replace \\[ and \\] by <script type="math/tex; mode=display">
    # and </script>, respectively.
    while True:
        i = s.find('\\[')
        if i == -1:
            break
        else:
            s = s[:i] + '<script type="math/tex; mode=display">' + s[i+2:]
            j = s.find('\\]')
            if j == -1:  # missing right-hand delimiter, so add one
                s = s + '</script>'
            else:
                s = s[:j] + '</script>' + s[j+2:]

    # Below t always has the "parsed so far" version of s, and s is
    # just the part of the original input s that hasn't been parsed.
    t = ''
    while True:
        i = s.find('$')
        if i == -1:
            # No dollar signs -- definitely done.
            return HtmlFragment(t + s)
        elif i > 0 and s[i-1] == '\\':
            # A dollar sign with a backslash right before it, so
            # we ignore it by sticking it in the parsed string t
            # and skip to the next iteration.
            t += s[:i-1] + '$'
            s = s[i+1:]
            continue
        elif i+1 < len(s) and s[i+1] == '$':
            # Found a math environment. Double dollar sign so display mode.
            disp = '; mode=display'
        else:
            # Found math environment. Single dollar sign so default mode.
            disp = ''

        # Now find the matching $ sign and form the html string.

        if len(disp) > 0:
            j = s[i+2:].find('$$')
            if j == -1:
                j = len(s)
                s += '$$'
            else:
                j += i + 2
            txt = s[i+2:j]
        else:
            j = s[i+2:].find('$')
            if j == -1:
                j = len(s)
                s += '$'
            else:
                j += i + 2
            txt = s[i+1:j]

        t += s[:i] + '<script type="math/tex%s">%s</script>'%(disp,
                      ' '.join(txt.splitlines()))
        s = s[j+1:]
        if len(disp) > 0:
            s = s[1:]
    return HtmlFragment(t)


class HTMLFragmentFactory(SageObject):

    def _repr_(self):
        """
        Return string representation

        OUTPUT:

        String.

        EXAMPLES::
        
            sage: html
            Create HTML output (see html? for details)
        """
        return 'Create HTML output (see html? for details)'
    
    def __call__(self, obj):
        r"""
        Construct a HTML fragment
     
        INPUT:
     
        - ``obj`` -- anything. An object for which you want a HTML
          representation.
     
        OUTPUT:
     
        A :class:`HtmlFragment` instance.
     
        EXAMPLES::
     
            sage: h = html('<hr>');  pretty_print(h)
            <hr>
            sage: type(h)
            <class 'sage.misc.html.HtmlFragment'>

            sage: pretty_print(html(1/2))
            <script type="math/tex">\frac{1}{2}</script>

            sage: pretty_print(html('<a href="http://sagemath.org">sagemath</a>'))
            <a href="http://sagemath.org">sagemath</a>
        """
        # Prefer dedicated _html_() method
        try:
            result = obj._html_()
        except AttributeError:
            pass
        else:
            if not isinstance(result, HtmlFragment):
                warnings.warn('_html_() did not return a HtmlFragment')
                return HtmlFragment(result)
            else:
                return result
        # Otherwise: convert latex to html
        try:
            result = obj._latex_()
        except AttributeError:
            pass
        else:
            return math_parse('${0}$'.format(obj._latex_()))
        # If all else fails
        return math_parse(str(obj))
         
    def eval(self, s, locals=None):
        r"""
        Evaluate embedded <sage> tags

        INPUT:

        - ``s`` -- string.

        - ``globals`` -- dictionary. The global variables when
          evaluating ``s``. Default: the current global variables.

        OUTPUT:
     
        A :class:`HtmlFragment` instance.
     
        EXAMPLES::

            sage: a = 123
            sage: html.eval('<sage>a</sage>')
            <script type="math/tex">123</script>
            sage: html.eval('<sage>a</sage>', locals={'a': 456})
            <script type="math/tex">456</script>
        """
        if locals is None:
            from sage.repl.user_globals import get_globals
            locals = get_globals()
        s = str(s)
        s = math_parse(s)
        t = ''
        while s:
            i = s.find('<sage>')
            if i == -1:
                 t += s
                 break
            j = s.find('</sage>')
            if j == -1:
                 t += s
                 break
            t += s[:i] + '<script type="math/tex">%s</script>'%\
                     latex(sage_eval(s[6+i:j], locals=locals))
            s = s[j+7:]
        return HtmlFragment(t)

    def iframe(self, url, height=400, width=800):
        r"""
        Generate an iframe HTML fragment

        INPUT:

        - ``url`` -- string. A url, either with or without URI scheme
          (defaults to "http"), or an absolute file path.

        - ``height`` -- the number of pixels for the page height.
          Defaults to 400.

        - ``width`` -- the number of pixels for the page width.
          Defaults to 800.

        OUTPUT:
     
        A :class:`HtmlFragment` instance.

        EXAMPLES::

            sage: pretty_print(html.iframe("sagemath.org"))
            <iframe height="400" width="800"
            src="http://sagemath.org"></iframe>
            sage: pretty_print(html.iframe("http://sagemath.org",30,40))
            <iframe height="30" width="40"
            src="http://sagemath.org"></iframe>
            sage: pretty_print(html.iframe("https://sagemath.org",30))
            <iframe height="30" width="800"
            src="https://sagemath.org"></iframe>
            sage: pretty_print(html.iframe("/home/admin/0/data/filename"))
            <iframe height="400" width="800"
            src="file:///home/admin/0/data/filename"></iframe>
            sage: pretty_print(html.iframe('data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAA'
            ....: 'AUAAAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBA'
            ....: 'AO9TXL0Y4OHwAAAABJRU5ErkJggg=="'))
            <iframe height="400" width="800" 
            src="http://data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAAO9TXL0Y4OHwAAAABJRU5ErkJggg==""></iframe>
        """
        if url.startswith('/'):
            url = 'file://{0}'.format(url)
        elif '://' not in url:
            url = 'http://{0}'.format(url)
        return HtmlFragment('<iframe height="{0}" width="{1}" src="{2}"></iframe>'
                            .format(height, width, url))


html = HTMLFragmentFactory()
html.__doc__ = HTMLFragmentFactory.__call__.__doc__



r"""
Elements of the algebra of differential forms

AUTHORS:

- Joris Vankerschaver (2010-07-25)

"""

# ****************************************************************************
#  Copyright (C) 2010 Joris Vankerschaver <joris.vankerschaver@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.symbolic.ring import SR
from sage.structure.element import RingElement, AlgebraElement
from sage.rings.integer import Integer
from sage.combinat.permutation import Permutation


def sort_subscript(subscript):
    """
    A subscript is a range of integers.  This function sorts a subscript
    in the sense of arranging it in ascending order.  The return values
    are the sign of the subscript and the sorted subscript, where the
    sign is defined as follows:

    #. sign == 0 if two or more entries in the subscript were equal.
    #. sign == +1, -1 if a positive (resp. negative) permutation was used to sort the subscript.


    INPUT:

    - ``subscript`` -- a subscript, i.e. a range of not necessarily
      distinct integers


    OUTPUT:

    - Sign of the permutation used to arrange the subscript, where 0
        means that the original subscript had two or more entries that
        were the same

    - Sorted subscript.


    EXAMPLES::

        sage: from sage.tensor.differential_form_element import sort_subscript
        sage: sort_subscript((1, 3, 2))
        (-1, (1, 2, 3))
        sage: sort_subscript((1, 3))
        (1, (1, 3))
        sage: sort_subscript((4, 2, 7, 9, 8))
        (1, (2, 4, 7, 8, 9))

    """
    if len(subscript) == 0:
        return 1, ()

    sub_list = sorted(subscript)
    offsets = [subscript.index(x)+1 for x in sub_list]

    # Check that offsets is a true permutation of 1..n
    n = len(offsets)
    if sum(offsets) != n*(n+1)//2:
        sign = 0
    else:
        sign = Permutation(offsets).signature()

    return sign, tuple(sub_list)



class DifferentialFormFormatter:
    r"""
    This class contains all the functionality to print a differential form in a
    graphically pleasing way.  This class is called by the ``_latex_`` and
    ``_repr_`` methods of the DifferentialForm class.

    In a nutshell (see the documentation of ``DifferentialForm`` for more
    details), differential forms are represented internally as a dictionary,
    where the keys are tuples representing the non-zero components of the
    form and the values are the component functions.  The methods of this
    class create string and latex representations out of the specification
    of a subscript and a component function.


    EXAMPLES::

        sage: from sage.tensor.differential_form_element import DifferentialFormFormatter
        sage: x, y, z = var('x, y, z')
        sage: U = CoordinatePatch((x, y, z))
        doctest:...: DeprecationWarning: Use Manifold instead.
        See http://trac.sagemath.org/24444 for details.
        sage: D = DifferentialFormFormatter(U)
        sage: D.repr((0, 2), sin(x*y))
        'sin(x*y)*dx/\\dz'
        sage: D.latex((0, 2), sin(x*y))
        '\\sin\\left(x y\\right) d x \\wedge d z'
        sage: D.latex((1, 2), exp(z))
        'e^{z} d y \\wedge d z'

    """
    def __init__(self, space):
        r"""
        Construct a differential form formatter.  See
        ``DifferentialFormFormatter`` for more information.

        INPUT:

        - space -- CoordinatePatch where the differential forms live.

        EXAMPLES::

            sage: from sage.tensor.differential_form_element import DifferentialFormFormatter
            sage: x, y, z = var('x, y, z')
            sage: U = CoordinatePatch((x, y, z))
            doctest:...: DeprecationWarning: Use Manifold instead.
            See http://trac.sagemath.org/24444 for details.
            sage: D = DifferentialFormFormatter(U)
            sage: D.repr((0, 2), sin(x*y))
            'sin(x*y)*dx/\\dz'

        """
        self._space = space


    def repr(self, comp, fun):
        r"""
        String representation of a primitive differential form, i.e. a function
        times a wedge product of d's of the coordinate functions.

        INPUT:

        - ``comp`` -- a subscript of a differential form.

        - ``fun`` -- the component function of this form.


        EXAMPLES::

            sage: from sage.tensor.differential_form_element import DifferentialFormFormatter
            sage: x, y, z = var('x, y, z')
            sage: U = CoordinatePatch((x, y, z))
            doctest:...: DeprecationWarning: Use Manifold instead.
            See http://trac.sagemath.org/24444 for details.
            sage: D = DifferentialFormFormatter(U)
            sage: D.repr((0, 1), z^3)
            'z^3*dx/\\dy'

        """

        str = "/\\".join( \
            [('d%r' % self._space.coordinate(c)) for c in comp])

        if fun == 1 and len(comp) > 0:
            # We have a non-trivial form whose component function is 1,
            # so we just return the formatted form part and ignore the 1.
            return str
        else:
            funstr = fun._repr_()

            if not self._is_atomic(funstr):
                funstr = '(' + funstr + ')'

            if len(str) > 0:
                return funstr + "*" + str
            else:
                return funstr


    def latex(self, comp, fun):
        r"""
        Latex representation of a primitive differential form, i.e. a function
        times a wedge product of d's of the coordinate functions.

        INPUT:

        - ``comp`` -- a subscript of a differential form.

        - ``fun`` -- the component function of this form.

        EXAMPLES::

            sage: from sage.tensor.differential_form_element import DifferentialFormFormatter
            sage: x, y, z = var('x, y, z')
            sage: U = CoordinatePatch((x, y, z))
            doctest:...: DeprecationWarning: Use Manifold instead.
            See http://trac.sagemath.org/24444 for details.
            sage: D = DifferentialFormFormatter(U)
            sage: D.latex((0, 1), z^3)
            'z^{3} d x \\wedge d y'
            sage: D.latex((), 1)
            '1'
            sage: D.latex((), z^3)
            'z^{3}'
            sage: D.latex((0,), 1)
            'd x'
        """

        from sage.misc.latex import latex

        s = " \\wedge ".join( \
            [('d %s' % latex(self._space.coordinate(c))) for c in comp])

        # Make sure this is a string and not a LatexExpr
        s = str(s)

        # Add coefficient except if it's 1
        if fun == 1:
            if s:
                return s
            else:
                return "1"

        funstr = fun._latex_()
        if not self._is_atomic(funstr):
            funstr = '(' + funstr + ')'

        if s:
            s = " " + s
        return funstr + s


    def _is_atomic(self, str):
        r"""
        Helper function to check whether a given string expression
        is atomic.

        EXAMPLES::

            sage: x, y, z = var('x, y, z')
            sage: U = CoordinatePatch((x, y, z))
            doctest:...: DeprecationWarning: Use Manifold instead.
            See http://trac.sagemath.org/24444 for details.
            sage: from sage.tensor.differential_form_element import DifferentialFormFormatter
            sage: D = DifferentialFormFormatter(U)
            sage: D._is_atomic('a + b')
            False
            sage: D._is_atomic('(a + b)')
            True
        """
        level = 0
        for n, c in enumerate(str):
            if c == '(':
                level += 1
            elif c == ')':
                level -= 1

            if c == '+' or c == '-':
                if level == 0 and n > 0:
                    return False
        return True



class DifferentialForm(AlgebraElement):
    r"""
    Differential form class.

    EXAMPLES:

    In order to instantiate differential forms of various degree, we begin
    by specifying the CoordinatePatch on which they live, as well as their
    parent DifferentialForms algebra.

    ::

        sage: x, y, z = var('x, y, z')
        sage: U = CoordinatePatch((x, y, z))
        doctest:...: DeprecationWarning: Use Manifold instead.
        See http://trac.sagemath.org/24444 for details.
        sage: F = DifferentialForms(U)
        doctest:...:  DeprecationWarning: For the set of differential forms of
         degree p, use U.diff_form_module(p), where U is the base manifold
         (type U.diff_form_module? for details).
        See http://trac.sagemath.org/24444 for details.
        sage: form1 = DifferentialForm(F, 0, sin(x*y)); form1
        doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
         where U is the base manifold (type U.diff_form? for details).
        See http://trac.sagemath.org/24444 for details.
        sin(x*y)

    In the previous example, we created a zero-form from a given function.
    To create forms of higher degree, we can use the subscript operator
    access the various components::

        sage: form2 = DifferentialForm(F, 1); form2
        0
        sage: form2[0] = 1
        sage: form2[1] = exp(cos(x))
        sage: form2[2] = 1/ln(y)
        sage: form2
        dx + e^cos(x)*dy + 1/log(y)*dz

    We may calculate the exterior derivative of a form, and observe that
    applying the exterior derivative twice always yields zero::

        sage: dform = form1.diff(); dform
        doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
         where U is the base manifold (type U.diff_form? for details).
        See http://trac.sagemath.org/24444 for details.
        y*cos(x*y)*dx + x*cos(x*y)*dy
        sage: dform.diff()
        0

    As can be seen from the previous example, the exterior derivative
    increases the degree of a form by one::

        sage: form2.degree()
        1
        sage: form2.diff().degree()
        2

    The ``d`` function provides a convenient shorthand for applying the
    diff member function.  Since d appears in other areas of mathematics
    as well, this function is not imported in the global namespace
    automatically::

        sage: from sage.tensor.differential_form_element import d
        sage: form2
        dx + e^cos(x)*dy + 1/log(y)*dz
        sage: d(form2)
        -e^cos(x)*sin(x)*dx/\dy + -1/(y*log(y)^2)*dy/\dz
        sage: form2.diff()
        -e^cos(x)*sin(x)*dx/\dy + -1/(y*log(y)^2)*dy/\dz
        sage: d(form1) == form1.diff()
        True

    The wedge product of two forms can be computed by means of the wedge
    member function::

        sage: form1 = DifferentialForm(F, 2)
        sage: form1[0, 1] = exp(z); form1
        e^z*dx/\dy
        sage: form2 = DifferentialForm(F, 1)
        sage: form2[2] = exp(-z)
        sage: form1.wedge(form2)
        doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
         where U is the base manifold (type U.diff_form? for details).
        See http://trac.sagemath.org/24444 for details.
        dx/\dy/\dz

    For this member function, there exists again a procedural function
    which is completely equivalent::

        sage: from sage.tensor.differential_form_element import wedge
        sage: form1.wedge(form2)
        dx/\dy/\dz
        sage: wedge(form1, form2)
        dx/\dy/\dz
        sage: form1.wedge(form2) == wedge(form1, form2)
        True


    .. NOTE::

        Differential forms are stored behind the screens as dictionaries,
        where the keys are the subscripts of the non-zero components, and
        the values are those components.

        For example, on a
        space with coordinates x, y, z, the form

            f = sin(x*y) dx /\\ dy + exp(z) dy /\\ dz

        would be represented as the dictionary

            {(0, 1): sin(x*y), (1, 2): exp(z)}.

        Most differential forms are ''sparse'' in the sense that most of
        their components are zero, so that this representation is more
        efficient than storing all of the components in a vector.

    """

    def __init__(self, parent, degree, fun = None):
        r"""
        Construct a differential form.

        INPUT:

        - ``parent`` -- Parent algebra of differential forms.

        - ``degree`` -- Degree of the differential form.

        - ``fun`` (default: None) -- Initialize this differential form with the given function.  If the degree is not zero, this argument is silently ignored.

        EXAMPLES::

            sage: x, y, z = var('x, y, z')
            sage: F = DifferentialForms(); F
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            doctest:...: DeprecationWarning: Use Manifold instead.
            See http://trac.sagemath.org/24444 for details.
            Algebra of differential forms in the variables x, y, z
            sage: f = DifferentialForm(F, 0, sin(z)); f
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sin(z)

        """

        from sage.tensor.differential_forms import DifferentialForms
        from sage.misc.superseded import deprecation
        deprecation(24444, 'Use U.diff_form(degree) instead, where U is ' +
                    'the base manifold (type U.diff_form? for details).')
        if not isinstance(parent, DifferentialForms):
            raise TypeError("Parent not an algebra of differential forms.")

        RingElement.__init__(self, parent)

        self._degree = degree
        self._components = {}

        if degree == 0 and fun is not None:
            self[[]] = fun

    def __getitem__(self, subscript):
        r"""
        Return a given component of the differential form.

        INPUT:

        - ``subscript`` -- subscript of the component.  Must be an integer
          or a list of integers.

        EXAMPLES::

            sage: x, y, z = var('x, y, z')
            sage: F = DifferentialForms(); F
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            Algebra of differential forms in the variables x, y, z
            sage: f = DifferentialForm(F, 0, sin(x*y)); f
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sin(x*y)
            sage: f[()]
            sin(x*y)

            sage: df = f.diff(); df
            y*cos(x*y)*dx + x*cos(x*y)*dy
            sage: df[0]
            y*cos(x*y)
            sage: df[1]
            x*cos(x*y)
            sage: df[2]
            0
        """
        if isinstance(subscript, (Integer, int)):
            subscript = (subscript, )
        else:
            subscript = tuple(subscript)

        dim = self.parent().base_space().dim()
        if any(s >= dim for s in subscript):
            raise ValueError("Index out of bounds.")

        if len(subscript) != self._degree:
            raise TypeError("%s is not a subscript of degree %s" %
                            (subscript, self._degree))

        sign, subscript = sort_subscript(subscript)

        if subscript in self._components:
            return sign*self._components[subscript]
        else:
            return 0

    def __setitem__(self, subscript, fun):
        r"""
        Modify a given component of the differential form.

        INPUT:

        - ``subscript`` -- subscript of the component.  Must be an integer
          or a list of integers.

        EXAMPLES::

            sage: F = DifferentialForms(); F
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            Algebra of differential forms in the variables x, y, z
            sage: f = DifferentialForm(F, 2)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f[1, 2] = x; f
            x*dy/\dz
        """

        if isinstance(subscript, (Integer, int)):
            subscript = (subscript, )
        else:
            subscript = tuple(subscript)

        dim = self.parent().base_space().dim()
        if any(s >= dim for s in subscript):
            raise ValueError("Index out of bounds.")

        if len(subscript) != self._degree:
            raise TypeError("%s is not a subscript of degree %s" %
                            (subscript, self._degree))

        sign, subscript = sort_subscript(subscript)
        self._components[subscript] = sign*SR(fun)


    def is_zero(self):
        r"""
        Return True if ``self`` is the zero form.

        EXAMPLES::

            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1); f
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            0
            sage: f.is_zero()
            True
            sage: f[1] = 1
            sage: f.is_zero()
            False
            sage: f.diff()
            0
            sage: f.diff().is_zero()
            True
        """

        self._cleanup()
        return len(self._components) == 0


    def degree(self):
        r"""
        Return the degree of self.

        EXAMPLES::

            sage: F = DifferentialForms(); F
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            Algebra of differential forms in the variables x, y, z
            sage: f = DifferentialForm(F, 2)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f[1, 2] = x; f
            x*dy/\dz
            sage: f.degree()
            2

        The exterior differential increases the degree of forms by one::

            sage: g = f.diff(); g
            dx/\dy/\dz
            sage: g.degree()
            3
        """
        return self._degree


    def __eq__(self, other):
        r"""
        Test whether two differential forms are equal.

        EXAMPLES::

            sage: F = DifferentialForms(); F
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            Algebra of differential forms in the variables x, y, z
            sage: f = DifferentialForm(F, 2)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f[1,2] = x; f
            x*dy/\dz
            sage: f == f
            True

            sage: g = DifferentialForm(F, 3)
            sage: g[0, 1, 2] = 1; g
            dx/\dy/\dz
            sage: f == g
            False
            sage: f.diff() == g
            True
        """
        if type(other) is type(self):
            if self._degree != other._degree:
                return False
            else:
                # TODO: the following two lines are where most of the
                # execution time is spent.
                self._cleanup()
                other._cleanup()

                if len(self._components) != len(other._components):
                    return False


                # We compare the component dictionary of both differential
                # forms, keeping in mind that the set of keys is
                # lexicographically ordered, so that we can simply iterate
                # over both dictionaries in one go and compare (key, value)
                # pairs as we go along.

                for (key1, val1), (key2, val2) in zip(self._components.items(),
                                                      other._components.items()):
                    if key1 != key2 or str(val1) != str(val2):
                        return False
                return True
        else:
            return False


    def __ne__(self, other):
        r"""
        Test whether two differential forms are not equal.

        EXAMPLES::

            sage: F = DifferentialForms(); F
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            Algebra of differential forms in the variables x, y, z
            sage: f = DifferentialForm(F, 2)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f[1,2] = x; f
            x*dy/\dz
            sage: g = DifferentialForm(F, 3)
            sage: g[0, 1, 2] = 1; g
            dx/\dy/\dz
            sage: f != g
            True
        """
        return not self == other

    def _neg_(self):
        r"""
        Return the negative of self.

        EXAMPLES::

            sage: x, y, z = var('x, y, z')
            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f[0] = y
            sage: f[1] = -x
            sage: f
            y*dx + -x*dy
            sage: -f
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            -y*dx + x*dy
            sage: -f == f._neg_()
            True
        """

        neg = DifferentialForm(self.parent(), self._degree)
        for comp in self._components:
            neg._components[comp] = -self._components[comp]
        return neg


    def _add_(self, other):
        r"""
        Add self and other

        EXAMPLES::

            sage: x, y, z = var('x, y, z')
            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: g = DifferentialForm(F, 1)
            sage: f[0] = exp(x); f
            e^x*dx
            sage: g[1] = sin(y); g
            sin(y)*dy
            sage: f + g
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            e^x*dx + sin(y)*dy
            sage: f + g == f._add_(g)
            True

        Forms must have the same degree to be added::

            sage: h = DifferentialForm(F, 2)
            sage: h[1, 2] = x; h
            x*dy/\dz
            sage: f + h
            Traceback (most recent call last):
            ...
            TypeError: Cannot add forms of degree 1 and 2

        Subtraction is implemented by adding the negative::

            sage: x, y, z = var('x, y, z')
            sage: F = DifferentialForms()
            sage: f = DifferentialForm(F, 1)
            sage: g = DifferentialForm(F, 1)
            sage: f[0] = exp(x); f
            e^x*dx
            sage: g[1] = sin(y); g
            sin(y)*dy
            sage: f - g
            e^x*dx + -sin(y)*dy
            sage: f - g == f._sub_(g)
            True

        Forms must have the same degree to be subtracted::

            sage: h = DifferentialForm(F, 2)
            sage: h[1, 2] = x; h
            x*dy/\dz
            sage: f - h
            Traceback (most recent call last):
            ...
            TypeError: Cannot add forms of degree 1 and 2
        """

        if self.is_zero():
            return other
        if other.is_zero():
            return self

        if self._degree != other._degree:
            raise TypeError("Cannot add forms of degree %s and %s" % \
                    (self._degree, other._degree))

        sumform = DifferentialForm(self.parent(), self._degree)
        sumform._components = self._components.copy()
        for comp, fun in other._components.items():
            sumform[comp] += fun

        sumform._cleanup()
        return sumform

    def _cleanup(self):
        r"""
        Helper function to clean up self, i.e. to remove any
        zero components from the dictionary of components.

        EXAMPLES::

            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f[0] = 0
            sage: f[1] = 1
            sage: f[2] = 0
            sage: f._dump_all()
            {(0,): 0, (1,): 1, (2,): 0}
            sage: f._cleanup()
            sage: f._dump_all()
            {(1,): 1}

        """
        zeros = []

        for comp in self._components:
            if self._components[comp].is_zero():
                zeros.append(comp)

        for comp in zeros:
            del self._components[comp]


    def _dump_all(self):
        r"""
        Helper function to dump the internal dictionary of form components.

        EXAMPLES::

            sage: x, y, z = var('x, y, z')
            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f[1] = exp(cos(x))
            sage: f[2] = sin(ln(y))
            sage: f
            e^cos(x)*dy + sin(log(y))*dz
            sage: f._dump_all()
            {(1,): e^cos(x), (2,): sin(log(y))}
            sage: g = DifferentialForm(F, 2)
            sage: g[1, 2] = x+y+z
            sage: g
            (x + y + z)*dy/\dz
            sage: g._dump_all()
            {(1, 2): x + y + z}

        """
        from pprint import pprint
        pprint(self._components)

    def diff(self):
        r"""
        Compute the exterior differential of ``self``.

        EXAMPLES::

            sage: x, y, z = var('x, y, z')
            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 0, sin(x*y)); f
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sin(x*y)
            sage: f.diff()
            y*cos(x*y)*dx + x*cos(x*y)*dy
            sage: g = DifferentialForm(F, 1)
            sage: g[0] = y/2
            sage: g[1] = -x/2
            sage: g
            1/2*y*dx + -1/2*x*dy
            sage: g.diff()
            -1*dx/\dy
            sage: h = DifferentialForm(F, 2)
            sage: h[0, 1] = exp(z)
            sage: h.diff()
            e^z*dx/\dy/\dz

        The square of the exterior differential operator is
        identically zero::

            sage: f
            sin(x*y)
            sage: f.diff()
            y*cos(x*y)*dx + x*cos(x*y)*dy
            sage: f.diff().diff()
            0

            sage: g.diff().diff()
            0

        The exterior differential operator is a derivation of degree one
        on the space of differential forms.  In this example we import the
        operator d() as a short-hand for having to call the diff()
        member function.

        ::

            sage: from sage.tensor.differential_form_element import d
            sage: d(f)
            y*cos(x*y)*dx + x*cos(x*y)*dy

            sage: d(f).wedge(g) + f.wedge(d(g))
            (-x*y*cos(x*y) - sin(x*y))*dx/\dy
            sage: d(f.wedge(g))
            (-x*y*cos(x*y) - sin(x*y))*dx/\dy

            sage: d(f.wedge(g)) == d(f).wedge(g) + f.wedge(d(g))
            True
        """

        diff_form = DifferentialForm(self.parent(), self._degree + 1)

        for comp in self._components:
            fun = self._components[comp]
            for n, coord in enumerate(self.parent().base_space().coordinates()):
                diff_form[(n, ) + comp] += fun.differentiate(coord)

        diff_form._cleanup()
        return diff_form


    def derivative(self, *args, **kwargs):
        r"""
        Compute the exterior derivative of ``self``.  This is the same as
        calling the ``diff`` member function.


        EXAMPLES::

            sage: x, y = var('x, y')
            sage: U = CoordinatePatch((x, y))
            doctest:...: DeprecationWarning: Use Manifold instead.
            See http://trac.sagemath.org/24444 for details.
            sage: F = DifferentialForms(U)
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: q = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: q[0] = -y/2
            sage: q[1] =  x/2
            sage: q.diff()
            dx/\dy
            sage: q.derivative()
            dx/\dy

        Invoking ``diff`` on a differential form has the same effect as
        calling this member function::

            sage: diff(q)
            dx/\dy
            sage: diff(q) == q.derivative()
            True

        When additional arguments are supplied to ``diff``, an error is raised,
        since only the exterior derivative has intrinsic meaning while
        derivatives with respect to the coordinate variables (in whichever
        way) are coordinate dependent, and hence not intrinsic.

        ::

            sage: diff(q, x)
            Traceback (most recent call last):
            ...
            ValueError: Differentiation of a form does not take any arguments.
        """

        if len(args) > 0 or len(kwargs) > 0:
            raise ValueError("Differentiation of a form does not take any arguments.")
        return self.diff()


    def wedge(self, other):
        r"""
        Returns the wedge product of ``self`` and other.

        EXAMPLES::

            sage: x, y, z = var('x, y, z')
            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f[0] = x^2
            sage: f[1] = y
            sage: f
            x^2*dx + y*dy
            sage: g = DifferentialForm(F, 1)
            sage: g[2] = z^3
            sage: g
            z^3*dz
            sage: f.wedge(g)
            x^2*z^3*dx/\dz + y*z^3*dy/\dz

        The wedge product is graded commutative::

            sage: f.wedge(g)
            x^2*z^3*dx/\dz + y*z^3*dy/\dz
            sage: g.wedge(f)
            -x^2*z^3*dx/\dz + -y*z^3*dy/\dz
            sage: f.wedge(f)
            0

        When the wedge product of forms belonging to different algebras
        is computed, an error is raised::

            sage: x, y, p, q = var('x, y, p, q')
            sage: F = DifferentialForms(CoordinatePatch((x, y)))
            doctest:...: DeprecationWarning: Use Manifold instead.
            See http://trac.sagemath.org/24444 for details.
            sage: G = DifferentialForms(CoordinatePatch((p, q)))
            sage: f = DifferentialForm(F, 0, 1); f
            1
            sage: g = DifferentialForm(G, 0, x); g
            x
            sage: f.parent()
            Algebra of differential forms in the variables x, y
            sage: g.parent()
            Algebra of differential forms in the variables p, q
            sage: f.wedge(g)
            Traceback (most recent call last):
            ...
            TypeError: unsupported operand parents for wedge: 'Algebra of differential forms in the variables x, y' and  'Algebra of differential forms in the variables p, q'

        """

        if self.parent() != other.parent():
            raise TypeError("unsupported operand parents for wedge: " +\
                "\'%s\' and  \'%s\'" % (self.parent(), other.parent()))

        output = DifferentialForm(self.parent(), self._degree + other._degree)
        if self._degree + other._degree > self.parent().ngens():
            return output

        for lcomp, lfun in self._components.items():
            for rcomp, rfun in other._components.items():
                output[lcomp + rcomp] += lfun*rfun

        output._cleanup()
        return output



    def _mul_(self, other):
        r"""
        Multiply self and other.  This is identical to the wedge operator.

        EXAMPLES::

            sage: x, y, z = var('x, y, z')
            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = F.gen(0); f
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            dx
            sage: g = F.gen(1); g
            dy
            sage: f*g
            dx/\dy
            sage: f.wedge(g)
            dx/\dy
            sage: f*g == f.wedge(g)
            True
            sage: f*g == f._mul_(g)
            True

        """
        return self.wedge(other)


    def _latex_(self):
        r"""
        Return a latex representation of self.

        EXAMPLES::

            sage: x, y, z = var('x, y, z')
            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f[1] = exp(z); f
            e^z*dy
            sage: latex(f)
            e^{z} d y
            sage: g = f.diff(); g
            -e^z*dy/\dz
            sage: latex(g)
            -e^{z} d y \wedge d z
            sage: latex(g) == g._latex_()
            True

        """
        if len(self._components) == 0:
            return '0'

        format = DifferentialFormFormatter(self.parent().base_space())
        output = [format.latex(comp, fun) \
                      for (comp, fun) in sorted(self._components.items())]
        return ' + '.join(output)


    def _repr_(self):
        r"""
        Return string representation of self.

        EXAMPLES::

            sage: x, y, z = var('x, y, z')
            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f[1] = exp(z); f
            e^z*dy
            sage: print(f)
            e^z*dy
            sage: f._repr_()
            'e^z*dy'
        """
        if len(self._components) == 0:
            return '0'

        format = DifferentialFormFormatter(self.parent().base_space())
        output = [format.repr(comp, fun) \
                      for (comp, fun) in sorted(self._components.items())]
        return ' + '.join(output)



    # Unsupported methods

    def abs(self):
        """
        Method not defined for differential forms.

        EXAMPLES::

            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f.abs()
            Traceback (most recent call last):
            ...
            NotImplementedError: Absolute value not defined for differential forms.

        """
        raise NotImplementedError("Absolute value not defined for differential forms.")


    def leading_coefficient(self, key=None):
        """
        Method not defined for differential forms.

        EXAMPLES::

            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f.leading_coefficient()
            Traceback (most recent call last):
            ...
            NotImplementedError: leading_coefficient not defined for differential forms.

        """
        raise NotImplementedError("leading_coefficient not defined for differential forms.")


    def leading_item(self, key=None):
        """
        Method not defined for differential forms.

        EXAMPLES::

            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f.leading_item()
            Traceback (most recent call last):
            ...
            NotImplementedError: leading_item not defined for differential forms.

        """
        raise NotImplementedError("leading_item not defined for differential forms.")


    def leading_monomial(self, key=None):
        """
        Method not defined for differential forms.

        EXAMPLES::

            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f.leading_monomial()
            Traceback (most recent call last):
            ...
            NotImplementedError: leading_monomial not defined for differential forms.

        """
        raise NotImplementedError("leading_monomial not defined for differential forms.")


    def leading_support(self, key=None):
        """
        Method not defined for differential forms.

        EXAMPLES::

            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f.leading_support()
            Traceback (most recent call last):
            ...
            NotImplementedError: leading_support not defined for differential forms.

        """
        raise NotImplementedError("leading_support not defined for differential forms.")


    def leading_term(self, key=None):
        """
        Method not defined for differential forms.

        EXAMPLES::

            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f.leading_term()
            Traceback (most recent call last):
            ...
            NotImplementedError: leading_term not defined for differential forms.

        """
        raise NotImplementedError("leading_term not defined for differential forms.")


    def trailing_coefficient(self, key=None):
        """
        Method not defined for differential forms.

        EXAMPLES::

            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f.trailing_coefficient()
            Traceback (most recent call last):
            ...
            NotImplementedError: trailing_coefficient not defined for differential forms.

        """
        raise NotImplementedError("trailing_coefficient not defined for differential forms.")


    def trailing_item(self, key=None):
        """
        Method not defined for differential forms.

        EXAMPLES::

            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f.trailing_item()
            Traceback (most recent call last):
            ...
            NotImplementedError: leading_coefficient not defined for differential forms.

        """
        raise NotImplementedError("leading_coefficient not defined for differential forms.")


    def trailing_monomial(self, key=None):
        """
        Method not defined for differential forms.

        EXAMPLES::

            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f.trailing_monomial()
            Traceback (most recent call last):
            ...
            NotImplementedError: trailing_monomial not defined for differential forms.

        """
        raise NotImplementedError("trailing_monomial not defined for differential forms.")


    def trailing_support(self, key=None):
        """
        Method not defined for differential forms.

        EXAMPLES::

            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f.trailing_support()
            Traceback (most recent call last):
            ...
            NotImplementedError: trailing_support not defined for differential forms.

        """
        raise NotImplementedError("trailing_support not defined for differential forms.")


    def trailing_term(self, key=None):
        """
        Method not defined for differential forms.

        EXAMPLES::

            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f.trailing_term()
            Traceback (most recent call last):
            ...
            NotImplementedError: trailing_term not defined for differential forms.

        """
        raise NotImplementedError("trailing_term not defined for differential forms.")


    def map_coefficients(self, f):
        """
        Method not defined for differential forms.

        EXAMPLES::

            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f.map_coefficients(lambda x: x)
            Traceback (most recent call last):
            ...
            NotImplementedError: map_coefficients not defined for differential forms.

        """
        raise NotImplementedError("map_coefficients not defined for differential forms.")


    def map_item(self, f):
        """
        Method not defined for differential forms.

        EXAMPLES::

            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f.map_item(lambda x: x)
            Traceback (most recent call last):
            ...
            NotImplementedError: map_item not defined for differential forms.

        """
        raise NotImplementedError("map_item not defined for differential forms.")


    def map_support(self, f):
        """
        Method not defined for differential forms.

        EXAMPLES::

            sage: F = DifferentialForms()
            doctest:...:  DeprecationWarning: For the set of differential forms of
             degree p, use U.diff_form_module(p), where U is the base manifold
             (type U.diff_form_module? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f = DifferentialForm(F, 1)
            doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
             where U is the base manifold (type U.diff_form? for details).
            See http://trac.sagemath.org/24444 for details.
            sage: f.map_support(lambda x: x)
            Traceback (most recent call last):
            ...
            NotImplementedError: map_support not defined for differential forms.

        """
        raise NotImplementedError("map_support not defined for differential forms.")




def d(form):
    r"""
    Returns the exterior derivative of a given form, i.e. calls the diff()
    member function.

    EXAMPLES::

        sage: from sage.tensor.differential_form_element import d
        sage: x, y, z = var('x, y, z')
        sage: F = DifferentialForms()
        doctest:...:  DeprecationWarning: For the set of differential forms of
         degree p, use U.diff_form_module(p), where U is the base manifold
         (type U.diff_form_module? for details).
        See http://trac.sagemath.org/24444 for details.
        sage: f = DifferentialForm(F, 1)
        doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
         where U is the base manifold (type U.diff_form? for details).
        See http://trac.sagemath.org/24444 for details.
        sage: f[2] = cos(x); f
        cos(x)*dz
        sage: d(f)
        -sin(x)*dx/\dz
        sage: f.diff()
        -sin(x)*dx/\dz
        sage: d(f) == f.diff()
        True
    """
    return form.diff()


def wedge(left, right):
    r"""
    Computes the wedge product of two forms, i.e. calls the wedge()
    member function.

    EXAMPLES::

        sage: from sage.tensor.differential_form_element import wedge
        sage: x, y, z = var('x, y, z')
        sage: F = DifferentialForms()
        doctest:...:  DeprecationWarning: For the set of differential forms of
         degree p, use U.diff_form_module(p), where U is the base manifold
         (type U.diff_form_module? for details).
        See http://trac.sagemath.org/24444 for details.
        sage: f = DifferentialForm(F, 1)
        doctest:...: DeprecationWarning: Use U.diff_form(degree) instead,
         where U is the base manifold (type U.diff_form? for details).
        See http://trac.sagemath.org/24444 for details.
        sage: f[2] = cos(x); f
        cos(x)*dz
        sage: g = DifferentialForm(F, 1)
        sage: g[1] = sin(y); g
        sin(y)*dy
        sage: wedge(f, g)
        -cos(x)*sin(y)*dy/\dz
        sage: f.wedge(g)
        -cos(x)*sin(y)*dy/\dz
        sage: wedge(f, g) == f.wedge(g)
        True
    """
    return left.wedge(right)
